import os
import traceback
from decimal import Decimal

import yaml
from rich.console import Console

DEFAULT_API_ENDPOINT = "https://api.app.predibase.com/v1"
DEFAULT_SERVING_ENDPOINT = "serving.app.predibase.com"

DEBUG = os.environ.get("PREDIBASE_DEBUG") is not None


_console = Console()
_error_console = Console(stderr=True, style="bold red")
_info_console = Console(stderr=True, style="bold blue")


def load_yaml(yaml_fp):
    with open(yaml_fp) as f:
        return yaml.safe_load(f)


def spinner(name):
    def decorator(fn):
        def wrap_fn(*args, **kwargs):
            with _console.status(f"{name}...", spinner="material"):
                try:
                    res = fn(*args, **kwargs)
                    _console.print(f"✅ {name}")
                    return res
                except Exception as e:
                    _console.print(f"💥 {name}")
                    _error_console.print(get_error_message(e))
                    raise

        return wrap_fn

    return decorator


def get_error_message(e: Exception) -> str:
    return f"{type(e).__name__}: {str(e)}" if not DEBUG else traceback.format_exc()


def log_info(v: str):
    _info_console.print(v)


def log_error(v: str):
    _error_console.print(v)


def get_url(session, endpoint: str) -> str:
    if "localhost" in session.url:
        root_url = "http://localhost:8000"
    else:
        url = session.url
        if "api." in url:
            url = url.replace("api.", "").replace("/v1", "")
        root_url = url
    from urllib.parse import urljoin

    return urljoin(root_url, endpoint)


class JSONFloat(float):
    def __repr__(self):
        return format(Decimal(self), "f")


class ConfigEncoder:
    def decimalize(self, val):
        if isinstance(val, dict):
            return {k: self.decimalize(v) for k, v in val.items()}

        if isinstance(val, (list, tuple)):
            return type(val)(self.decimalize(v) for v in val)

        if isinstance(val, float):
            return JSONFloat(val)

        return val



from typing import Optional

import typer
from rich import box
from rich.progress import Progress, SpinnerColumn, TextColumn
from rich.table import Table

from predibase.cli_commands.utils import df_to_table, get_client, get_console
from predibase.resource.llm.response import GeneratedResponse

app = typer.Typer(no_args_is_help=True)


@app.command(help="Query a Large Language Model (LLM)")
def llm(
    template: str = typer.Option(
        None,
        "--template",
        "-t",
        prompt="Template input",
        prompt_required=True,
        help="Prompt template input text",
    ),
    deployment_name: str = typer.Option(
        None,
        "--deployment-name",
        "-m",
        prompt="Deployment to prompt",
        prompt_required=True,
        help="The name of the deployment to prompt",
    ),
    index_name: Optional[str] = typer.Option(
        None,
        "--index-name",
        "-i",
        prompt_required=False,
        help="Optional dataset to index",
    ),
    dataset_name: Optional[str] = typer.Option(
        None,
        "--dataset-name",
        "-d",
        prompt_required=False,
        help="Optional dataset for batch inference",
    ),
    limit: Optional[int] = typer.Option(None, "--limit", "-l", prompt_required=False, help="Optional results limit"),
):
    client = get_client()

    with Progress(
        SpinnerColumn(),
        TextColumn("[progress.description]{task.description}"),
        transient=True,
    ) as progress:
        progress.add_task(description="Querying LLM...", total=None)
        responses = client.LLM(f"pb://deployments/{deployment_name}").prompt(
            template,
            # return_df=True,
        )
        df = GeneratedResponse.to_pandas(responses)

    table = Table(show_header=True, header_style="bold magenta")

    # Modify the table instance to have the data from the DataFrame
    table = df_to_table(df, table)

    # Update the style of the table
    table.row_styles = ["none", "dim"]
    table.box = box.SIMPLE_HEAD

    get_console().print(table)


if __name__ == "__main__":
    app()



from typing import Dict


def max_epochs_of_model_response(model: Dict):
    if "modelRun" in model and "renderedConfig" in model["modelRun"]:
        return model["modelRun"]["renderedConfig"]["trainer"]["epochs"]
    return None



from typing import Optional

import typer

from predibase.cli_commands.utils import get_client

app = typer.Typer(no_args_is_help=True)


@app.command(help="Create a Large Language Model (LLM) deployment")
def model(
    repo_name: str = typer.Option(
        None,
        "--repo-name",
        "--model-name",
        "-r",
        "-m",
        prompt="Name of the model",
        prompt_required=True,
        help="Name of the model or model repo",
    ),
    version: Optional[int] = typer.Option(
        None,
        "--version",
        "-v",
        prompt="Model version",
        prompt_required=False,
        help="Optionally specify the exact version to download, otherwise defaults to latest",
    ),
    file_name: Optional[str] = typer.Option(
        None,
        "--name",
        "-n",
        prompt="Name of the downloaded file",
        prompt_required=False,
        help="Name of the downloaded file",
    ),
    location: Optional[str] = typer.Option(
        None,
        "--location",
        "-l",
        prompt="Path to save the downloaded file to",
        prompt_required=False,
        help="Path to save the downloaded file to",
    ),
):
    client = get_client()
    m = client.get_model(repo_name, version)
    m.download(name=file_name, location=location)


if __name__ == "__main__":
    app()



import concurrent.futures
from abc import abstractmethod
from typing import Dict, List, Optional, Union

import deprecation
import pandas as pd

from predibase.pql.api import Session
from predibase.resource.connection import Connection
from predibase.resource.dataset import Dataset
from predibase.resource.llm.interface import GeneratedResponse, HuggingFaceLLM, LLMDeployment, LLMDeploymentJob
from predibase.version import __version__


class LlmMixin:
    session: Session

    def LLM(self, uri: str) -> Union[HuggingFaceLLM, LLMDeployment]:
        if uri.startswith("pb://deployments/"):
            return LLMDeployment(self.session, uri[len("pb://deployments/") :])

        if uri.startswith("hf://"):
            return HuggingFaceLLM(self.session, uri[len("hf://") :])

        raise ValueError(
            "must provide either a Hugging Face URI (hf://<...>) "
            "or a Predibase deployments URI (pb://deployments/<name>).",
        )

    def generate(
        self,
        templates: Union[str, List[str]],
        deployment_names: Union[str, List[str]],
        options: Optional[Dict[str, float]] = None,
    ) -> pd.DataFrame:
        deployment_names = [deployment_names] if isinstance(deployment_names, str) else deployment_names
        templates = [templates] if isinstance(templates, str) else templates

        resp_list, future_to_args = [], dict()
        with concurrent.futures.ThreadPoolExecutor() as executor:
            futures = []
            for input_text in templates:
                for deployment_name in deployment_names:
                    future = executor.submit(
                        self.session.post_json,
                        f"/llms/{deployment_name}/generate",
                        {
                            "inputs": input_text,
                            "parameters": options,
                        },
                    )
                    future_to_args[future] = (deployment_name, input_text)
                    futures.append(future)

            for future in concurrent.futures.as_completed(futures):
                try:
                    deployment_name, input_text = future_to_args[future]
                    res = future.result()
                    res = GeneratedResponse(
                        prompt=input_text,
                        response=res["generated_text"],
                        model_name=deployment_name,
                        generated_tokens=res["details"]["generated_tokens"],
                    )
                    resp_list.append(res)
                except Exception as exc:
                    print("ERROR:", exc)
        return GeneratedResponse.to_pandas(resp_list)

    @deprecation.deprecated(
        deprecated_in="2023.5.8",
        current_version=__version__,
        details="Use the LLMDeployment().prompt method instead",
    )
    def prompt(
        self,
        templates: Union[str, List[str]],
        deployment_names: Union[str, List[str]],
        index: Optional[Union[str, Dataset]] = None,
        dataset: Optional[Union[str, Dataset]] = None,
        limit: Optional[int] = 10,
        options: Optional[Dict[str, float]] = None,
    ) -> pd.DataFrame:
        if not (index or dataset):
            # talk directly to the LLM, bypassing Temporal and the engines.
            return self.generate(templates=templates, deployment_names=deployment_names, options=options)

        # Set the connection if we have a dataset or index
        conn_id = None
        if isinstance(dataset, Dataset):
            conn_id = dataset.connection_id
        elif isinstance(index, Dataset):
            conn_id = index.connection_id
        resp = self.session.post_json(
            "/prompt",
            json={
                "connectionID": conn_id,
                "deploymentNames": [deployment_names] if isinstance(deployment_names, str) else deployment_names,
                "templates": [templates] if isinstance(templates, str) else templates,
                "options": options,
                "limit": limit,
                "indexName": index.name if isinstance(index, Dataset) else index,
                "datasetName": dataset.name if isinstance(dataset, Dataset) else dataset,
            },
            timeout=300,  # increase timeout to 5 minutes for this request
        )
        responses = [GeneratedResponse(**r) for r in resp["responses"]]
        return GeneratedResponse.to_pandas(responses)

    @deprecation.deprecated(
        deprecated_in="2023.5.8",
        current_version=__version__,
        details="Use the HuggingFaceLLM().deploy method instead",
    )
    def deploy_llm(
        self,
        deployment_name: str,
        model_name: str,
        engine_template: Optional[str] = None,
        hf_token: Optional[str] = None,
        auto_suspend_seconds=None,
    ) -> LLMDeploymentJob:
        return self.LLM(model_name).deploy(
            deployment_name,
            engine_template=engine_template,
            auto_suspend_seconds=auto_suspend_seconds,
            hf_token=hf_token,
        )

    @deprecation.deprecated(
        deprecated_in="2023.5.8",
        current_version=__version__,
        details="Use the LLMDeployment().delete method instead",
    )
    def delete_llm(self, deployment_name: str):
        self.session.delete_json(f"/llms/{deployment_name}")

    def list_all_llms(self):
        return self.session.get_json("/llms?activeOnly=false")

    def list_deployed_llms(self):
        return self.session.get_json("/llms")

    @abstractmethod
    def get_dataset(self) -> Dataset:
        pass

    @abstractmethod
    def list_connections(self) -> List[Connection]:
        pass



import os
from dataclasses import dataclass
from typing import Any, Dict, Optional

import pandas as pd
import yaml
from rich.console import Console
from rich.table import Table

from predibase import PredibaseClient
from predibase.pql.api import Session
from predibase.resource.dataset import Dataset
from predibase.resource.engine import Engine
from predibase.resource.model import ModelRepo
from predibase.util import DEFAULT_API_ENDPOINT


@dataclass
class Defaults:
    repo: Optional[str] = None
    engine: Optional[str] = None
    session: Optional[Session] = None


console = Console()
defaults: Defaults = Defaults()


def get_console() -> Console:
    return console


def get_client() -> PredibaseClient:
    return PredibaseClient(session=defaults.session)


def sanitize_engine_name(name: str) -> str:
    return name.replace("/", "-")


def get_or_create_repo(repo_name: Optional[str] = None) -> ModelRepo:
    repo_name = repo_name or defaults.repo
    if repo_name is None:
        raise ValueError("Repo name is required")
    return get_client().create_model_repo(name=repo_name, exists_ok=True)


def get_repo(repo_name: Optional[str] = None) -> ModelRepo:
    repo_name = repo_name or defaults.repo
    if repo_name is None:
        raise ValueError("Repo name is required")
    return get_client().get_model_repo(name=repo_name)


def get_engine(engine_name: Optional[str] = None) -> Engine:
    engine_name = engine_name or defaults.engine
    if engine_name is None:
        raise ValueError("Engine name is required")
    return get_client().get_engine(name=engine_name)


def get_dataset(dataset_name: Optional[str] = None) -> Dataset:
    if dataset_name is None:
        raise ValueError("Dataset name is required")

    conn_name = None
    if "/" in dataset_name:
        conn_name, dataset_name = dataset_name.split("/")
    return get_client().get_dataset(dataset_name=dataset_name, connection_name=conn_name)


def set_defaults_from_settings(settings: Dict[str, Any]):
    url = settings.get("endpoint")
    if url is None:
        url = os.environ.get("PREDIBASE_GATEWAY", DEFAULT_API_ENDPOINT)

    global defaults
    defaults = Defaults(
        repo=settings.get("repo"),
        engine=settings.get("engine"),
        session=Session(token=settings.get("token"), url=url),
    )


def load_yaml(fname: str) -> Dict[str, Any]:
    with open(fname) as f:
        return yaml.safe_load(f)


def df_to_table(
    pandas_dataframe: pd.DataFrame,
    rich_table: Table,
    show_index: bool = True,
    index_name: Optional[str] = None,
) -> Table:
    """Convert a pandas.DataFrame obj into a rich.Table obj.

    Args:
        pandas_dataframe (DataFrame): A Pandas DataFrame to be converted to a rich Table.
        rich_table (Table): A rich Table that should be populated by the DataFrame values.
        show_index (bool): Add a column with a row count to the table. Defaults to True.
        index_name (str, optional): The column name to give to the index column. Defaults to None, showing no value.
    Returns:
        Table: The rich Table instance passed, populated with the DataFrame values.
    """

    if show_index:
        index_name = str(index_name) if index_name else ""
        rich_table.add_column(index_name)

    for column in pandas_dataframe.columns:
        rich_table.add_column(str(column))

    for index, value_list in enumerate(pandas_dataframe.values.tolist()):
        row = [str(index)] if show_index else []
        row += [str(x) for x in value_list]
        rich_table.add_row(*row)

    return rich_table



from typing import Any, Dict, List, Optional, Union

import pandas as pd

from predibase.resource.deployment import Deployment
from predibase.resource.model import Model
from predibase.util import spinner


class DeploymentMixin:
    @spinner(name="Create Deployment")
    def create_deployment(
        self,
        name: str,
        model: Model,
        engine_name: Optional[str] = None,
        replace: bool = False,
        exists_ok: bool = False,
        comment: Optional[str] = None,
        df: bool = False,
    ) -> Union[pd.DataFrame, List[Deployment]]:
        """Creates a deployment.

        :param str name: Name of deployment.
        :param Model model: Model object to deploy.
        :param str engine_name: Optional serving engine name to deploy to (Default None).
        :param bool replace: Optional flag replace a deployment (default False).
        :param bool exists_ok: Optional flag to only create deployment if not exists (default False).
        :param str comment: Optional comment for the deployment (default None).
        :param bool df: Optional flag to return the result as a dataframe (Default False).
        :return: pandas DataFrame or list of Deployment objects.
        """
        if self.session.is_free_trial():
            raise PermissionError(
                "Deployments are locked during the trial period. Contact us to upgrade or if you would like a demo",
            )
        elif self.session.is_plan_expired():
            raise PermissionError(
                "Deployments are locked for expired plans. Contact us to upgrade or if you would like a demo",
            )
        else:
            conditions = []
            if replace:
                conditions.append("OR REPLACE")
            conditions.append("DEPLOYMENT")
            if exists_ok:
                conditions.append("IF NOT EXISTS")
            conditions.append(f'"{name}"')

            if engine_name is not None:
                conditions.append(f'TO "{engine_name}"')
            conditions.append(f'USING "{model.repo.name}" VERSION {model.version}')
            if comment is not None:
                conditions.append(f"COMMENT '{comment}'")

            query = f"CREATE {' '.join(conditions)}"
            result = self.session.execute(query)
            deployments = self._format_deployment_result(result)
            # TODO: Change this to return a DeploymentCollection, with a to_dataframe method.
            if df:
                return pd.DataFrame(deployments)
            return [Deployment.from_dict({"session": self.session, **x}) for x in deployments]

    @spinner(name="Delete Deployment")
    def delete_deployment(
        self,
        name: str,
        if_exists: bool = False,
        df: bool = False,
    ) -> Union[pd.DataFrame, List[Deployment]]:
        """Deletes a deployment.

        :param str name: required name of deployment.
        :param bool if_exists: Optional flag to only delete deployment if it exists (default False).
        :param bool df: Optional flag to return the result as a dataframe (Default False).
        :return: pandas DataFrame or list of Deployment objects.
        """
        conditions = []
        conditions.append("DROP DEPLOYMENT")
        if if_exists:
            conditions.append("IF EXISTS")
        conditions.append(f'"{name}"')

        query = f"{' '.join(conditions)}"
        result = self.session.execute(query)
        deployments = self._format_deployment_result(result)
        if df:
            return pd.DataFrame(deployments)
        return [Deployment.from_dict({"session": self.session, **x}) for x in deployments]

    def list_deployments(
        self,
        deployment_name_pattern: Optional[str] = None,
        engine_name: Optional[str] = None,
        repo_name: Optional[str] = None,
        model_version: Optional[str] = None,
        df: bool = False,
    ) -> Union[pd.DataFrame, List[Deployment]]:
        """Lists Deployments.

        :param str deployment_name_pattern: Optional filter for deployment name (Default None).
        :param str engine_name: Optional filter by engine name (Default None).
        :param str repo_name: Optional filter by model repository name (Default None).
        :param str model_version: Optional filter by model version (Default None).
        :param bool df: Optional flag to return the result as a dataframe (Default False).
        :return: pandas DataFrame or list of Deployment objects.
        """
        conditions = []
        if engine_name is not None:
            conditions.append(f'IN "{engine_name}"')
        if repo_name is not None:
            conditions.append(f'USING "{repo_name}"')
            if model_version is not None:
                conditions.append(f"VERSION {model_version}")
        if deployment_name_pattern is not None:
            conditions.append(f"LIKE '{deployment_name_pattern}'")

        query = f"SHOW DEPLOYMENTS {' '.join(conditions)}"
        result = self.session.execute(query)
        deployments = self._format_deployment_result(result)
        if df:
            return pd.DataFrame(deployments)
        return [Deployment.from_dict({"session": self.session, **x}) for x in deployments]

    def _format_deployment_result(self, result: pd.DataFrame) -> List[Dict[str, Any]]:
        deployments = [
            {
                "session": self.session,
                "name": row["name"],
                "deploymentUrl": row["url"],
                "deploymentVersion": row["head_version_number"],
                "engineName": row["head_version_engine_name"],
                "modelName": row["head_version_model_name"],
                "modelVersion": row["head_version_model_version"],
                "comment": row["head_version_comment"],
                "errorText": row["head_version_error_text"],
            }
            for row in result.to_dict(orient="records")
        ]
        return deployments

    def get_deployment(self, name: str) -> Deployment:
        """Gets a deployment.

        :param str name: Name of deployment.
        :return: Deployment object.
        """
        deployments = self.list_deployments(deployment_name_pattern=name, df=False)
        if len(deployments) == 0:
            raise ValueError(f"Deployment '{name}' not found.")
        return deployments[0]



import concurrent.futures
import os
import time
from dataclasses import dataclass, field
from typing import Dict, List, Optional, TYPE_CHECKING, Union

import yaml
from dataclasses_json import config, dataclass_json, LetterCase

from predibase.pql.api import ServerResponseError
from predibase.resource.connection import get_dataset
from predibase.resource.dataset import Dataset
from predibase.resource.engine import Engine
from predibase.resource.llm.response import GeneratedResponse
from predibase.resource.model import create_model_repo, ModelFuture, ModelRepo
from predibase.util import load_yaml

if TYPE_CHECKING:
    from predibase.pql.api import Session


_PATH_HERE = os.path.abspath(os.path.dirname(__file__))
_TEMPLATE_DIR = os.path.join(_PATH_HERE, "templates")


class HuggingFaceLLM:
    def __init__(self, session: "Session", model_name: str):
        self.session = session
        self.model_name = model_name

    def deploy(
        self,
        deployment_name: str,
        engine_template: Optional[str] = None,
        hf_token: Optional[str] = None,
        auto_suspend_seconds: Optional[int] = None,
    ) -> "LLMDeploymentJob":
        params = {"modelName": self.model_name}
        if hf_token:
            params["hfToken"] = hf_token
        data = self.session.get_json("/supported_llms", params=params)
        llm_config = data["deploymentSpec"]

        message = data["message"]
        if message:
            print("=" * len(message))
            print(message)
            print("=" * len(message))

        model_params = {
            **llm_config,
            "name": deployment_name,
            "modelName": self.model_name,
            "engineTemplate": engine_template,
            "scaleDownPeriod": auto_suspend_seconds,
        }

        if hf_token is not None:
            model_params["hfToken"] = hf_token

        print("Deploying the model with the following params:")
        print(model_params)

        self.session.post_json(
            "/llms",
            json=model_params,
        )

        return LLMDeploymentJob(deployment_name, self.session)

    def finetune(
        self,
        template: Optional[str] = None,
        target: Optional[str] = None,
        dataset: Optional[Union[str, Dataset]] = None,
        engine: Optional[Union[str, Engine]] = None,
        config: Optional[Union[str, Dict]] = None,
        repo: Optional[str] = None,
    ) -> ModelFuture:
        if config is None:
            with open(os.path.join(_TEMPLATE_DIR, "defaults.yaml")) as f:
                config_str = f.read()
            config_str = config_str.format(base_model=self.model_name, template=template, target=target)
            print(config_str)
            config = yaml.safe_load(config_str)
        else:
            config = load_yaml(config)

        if repo is None:
            dataset_name = dataset.name if isinstance(dataset, Dataset) else dataset
            if "/" in dataset_name:
                _, dataset_name = dataset_name.split("/")

            model_name = self.model_name
            if "/" in model_name:
                _, model_name = model_name.split("/")

            repo = f"{model_name}-{dataset_name}"

        if "/" in repo:
            repo = repo.replace("/", "-")

        repo: ModelRepo = get_or_create_repo(self.session, repo)
        if dataset is None:
            # Assume the dataset is the same as the repo head
            md = repo.head().to_draft()
            md.config = config
        else:
            if isinstance(dataset, str):
                conn_name = None
                if "/" in dataset:
                    conn_name, dataset = dataset.split("/")
                dataset = get_dataset(self.session, dataset, connection_name=conn_name)
            md = repo.create_draft(config=config, dataset=dataset)

        return md.train_async(engine=engine)


@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclass
class _LLMDeployment:
    id: int = field(metadata=config(field_name="id"))
    tenant_id: int = field(metadata=config(field_name="tenantID"))
    uuid: str = field(metadata=config(field_name="uuid"))
    name: str = field(metadata=config(field_name="name"))
    description: Optional[str] = field(metadata=config(field_name="description"))
    model_name: str = field(metadata=config(field_name="modelName"))
    num_shards: Optional[int] = field(metadata=config(field_name="numShards"))
    quantize: bool = field(metadata=config(field_name="quantize"))
    deployment_status: str = field(metadata=config(field_name="deploymentStatus"))

    prompt_template: str = field(metadata=config(field_name="promptTemplate"))
    min_replicas: int = field(metadata=config(field_name="minReplicas"))
    max_replicas: int = field(metadata=config(field_name="maxReplicas"))
    created: str = field(metadata=config(field_name="created"))
    updated: str = field(metadata=config(field_name="updated"))
    created_by_user_id: Optional[int] = field(metadata=config(field_name="createdByUserID"))
    scale_down_period: int = field(metadata=config(field_name="scaleDownPeriod"))
    error_text: Optional[str] = field(metadata=config(field_name="errorText"), default=None)


@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclass
class LLMDeploymentReadyResponse:
    name: str
    ready: bool


@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclass
class LLMDeploymentScaledResponse:
    name: str
    scaled: bool


class LLMDeployment:
    def __init__(self, session: "Session", name: str, deployment_metadata: Optional[_LLMDeployment] = None):
        self.session = session
        self.name = name
        self.data = (
            _LLMDeployment.from_dict(self.session.get_json(f"/llms/{self.name}"))
            if deployment_metadata is None
            else deployment_metadata
        )

    def generate(
        self,
        templates: Union[str, List[str]],
        options: Optional[Dict[str, float]] = None,
    ) -> List[GeneratedResponse]:
        templates = [templates] if isinstance(templates, str) else templates

        resp_list, future_to_args = [], dict()
        with concurrent.futures.ThreadPoolExecutor() as executor:
            futures = []
            for input_text in templates:
                future = executor.submit(
                    self.session.post_json,
                    f"/llms/{self.name}/generate",
                    {
                        "inputs": input_text,
                        "parameters": options,
                    },
                )
                future_to_args[future] = (self.name, input_text)
                futures.append(future)

            for future in concurrent.futures.as_completed(futures):
                try:
                    deployment_name, input_text = future_to_args[future]
                    res = future.result()
                    res = GeneratedResponse(
                        prompt=input_text,
                        response=res["generated_text"],
                        model_name=deployment_name,
                        generated_tokens=res["details"]["generated_tokens"],
                    )
                    resp_list.append(res)
                except Exception as exc:
                    print("ERROR:", exc)
        return resp_list

    def prompt(
        self,
        templates: Union[str, List[str]],
        options: Optional[Dict[str, float]] = None,
    ) -> List[GeneratedResponse]:
        deployment_status = self.get_status()
        if deployment_status != "active":
            raise Exception(f"Target LLM deployment `{self.name}` is not active yet!")

        deployment_ready = self.is_ready()
        if not deployment_ready:
            print(f"WARNING: Target LLM deployment `{self.name}` is not fully scaled yet. Responses may be delayed...")

        # talk directly to the LLM, bypassing Temporal and the engines.
        return self.generate(templates=templates, options=options)

    def delete(self):
        print(f"Requested deletion of llm deployment: `{self.name}` ...")
        endpoint = f"/llms/{self.name}"
        resp = self.session._delete(endpoint)
        if not resp.ok:
            raise RuntimeError(f"Failed to trigger LLM deletion - got status {resp.status_code} -- {resp.reason}")

        while True:
            try:
                self.session.get_json(endpoint)
            except Exception as e:
                if isinstance(e, ServerResponseError) and "record not found" in e.message:
                    print(f"Successfully deleted llm deployment: `{self.name}`")
                    break
                else:
                    raise RuntimeError(f"Error while deleting deployment `{self.name}`: {e} {type(e)}")
            time.sleep(1.0)

    def is_ready(self) -> bool:
        resp = self.session.get_json(f"/llms/{self.name}/ready")
        return LLMDeploymentReadyResponse.from_dict(resp).ready

    def wait_for_ready(self, timeout_seconds: int = 600, poll_interval_seconds: int = 5) -> bool:
        start = time.time()
        while int(time.time() - start) < timeout_seconds:
            if self.is_ready():
                return True
            time.sleep(poll_interval_seconds)
        return False

    def is_scaled(self) -> bool:
        resp = self.session.get_json(f"/llms/{self.name}/scaled")
        return LLMDeploymentScaledResponse.from_dict(resp).scaled

    def get_status(self) -> str:
        resp = _LLMDeployment.from_dict(self.session.get_json(f"/llms/{self.name}"))
        return resp.deployment_status


class LLMDeploymentJob:
    def __init__(self, deployment_name: str, session: "Session"):
        self._deployment_name = deployment_name
        self._uri = f"pb://jobs/deploy::{deployment_name}"
        self._session = session

    def get(self) -> LLMDeployment:
        resp = self._session.get_llm_deployment_until_with_logging(
            f"/llms/{self._deployment_name}",
            lambda resp: resp["deploymentStatus"] == "active",
            lambda resp: f"Failed to create deployment {self._deployment_name} with status {resp['deploymentStatus']}"
            if resp["deploymentStatus"] in ("failed", "canceled")
            else None,
        )

        return LLMDeployment(self._session, self._deployment_name, resp)

    def cancel(self):
        return self._session.post_json(f"/llms/{self._deployment_name}/cancel", {})


def get_or_create_repo(session: "Session", repo_name: str) -> ModelRepo:
    return create_model_repo(session, name=repo_name, exists_ok=True)



import re
from typing import Optional

import typer

from predibase.cli_commands.utils import get_client, get_console

app = typer.Typer(no_args_is_help=True)


@app.command(help="Create a Large Language Model (LLM) deployment")
def llm(
    deployment_name: str = typer.Option(
        None,
        "--deployment-name",
        "-d",
        prompt="Deployment name",
        prompt_required=True,
        help="Name of the deployment",
    ),
    model_name: str = typer.Option(
        None,
        "--model-name",
        "-m",
        prompt="Model to deploy",
        prompt_required=True,
        help="Name of the model, e.g., hf://meta-llama/Llama-2-7b-hf",
    ),
    engine_template: Optional[str] = typer.Option(
        None,
        "--engine-template",
        "-e",
        prompt="Engine name",
        prompt_required=False,
        help="Optional engine template to provision for hosting the model",
    ),
    auto_suspend_secs: Optional[int] = typer.Option(
        None,
        "--auto-suspend",
        "-s",
        prompt="Auto suspend (seconds)",
        prompt_required=False,
        help="Optional auto suspend time in seconds.  Set to zero to disable auto scaling down.",
    ),
    hf_token: Optional[str] = typer.Option(
        None,
        "--hf-token",
        prompt="HuggingFace API token",
        prompt_required=False,
        help="Optional HuggingFace API token for private models",
    ),
    wait: Optional[bool] = typer.Option(
        None,
        "--wait",
        prompt="Whether to wait until deployment finishes",
        prompt_required=False,
        help="If set, the deploy command will not return until the deployment process finishes",
    ),
):
    # raise ValueError if name is not lower case alphanumeric characters or '-'
    if re.match(r"^[a-z0-9-]+$", deployment_name) is None:
        raise ValueError("name must be lower case alphanumeric characters or '-'")

    client = get_client()

    get_console().print("Deploying an LLM with the following parameters:")
    get_console().print("\tdeployment_name:", deployment_name)
    get_console().print("\tmodel_name:", model_name)
    if engine_template:
        get_console().print("\tengine_template:", engine_template)
    if auto_suspend_secs:
        get_console().print("\tauto_suspend_seconds:", auto_suspend_secs)

    job = client.LLM(model_name).deploy(
        deployment_name,
        engine_template=engine_template,
        auto_suspend_seconds=auto_suspend_secs,
        hf_token=hf_token,
    )
    get_console().print("Deploy request sent.")

    if wait:
        get_console().print("Waiting for deploy to complete...")
        job.get()
        get_console().print("Deploy process finished.")


if __name__ == "__main__":
    app()



import copy
from concurrent.futures import as_completed, ThreadPoolExecutor
from typing import Any, Dict, List, Optional, Union

import pandas as pd

from predibase.pql.api import ServerResponseError
from predibase.resource.config import ConfigSuggestion
from predibase.resource.dataset import Dataset
from predibase.resource.engine import Engine
from predibase.resource.model import create_model_repo
from predibase.resource.model import get_default_config as _get_default_config
from predibase.resource.model import get_model as _get_model
from predibase.resource.model import get_model_repo
from predibase.resource.model import get_suggested_configs as _get_suggested_configs
from predibase.resource.model import Model, ModelRepo
from predibase.util import ConfigEncoder, get_url, load_yaml, log_info
from predibase.util.url import encode_url_param


class ModelMixin:
    def list_model_repos(
        self,
        name: str = None,
        description: str = None,
        df: bool = False,
        limit: Optional[int] = 999999,
    ) -> Union[pd.DataFrame, List[ModelRepo]]:
        """
        :param name: Filter by name
        :param description: Filter by description
        :param df: Whether to return pandas dataframe or list of objects
        :param limit: max number of model repos to return
        :return: Pandas dataframe or list of model repos
        """
        endpoint = f"/models?limit={limit}"
        if name:
            endpoint += f"&searchKeys[]=modelName&searchVals[]={name}"
        if description:
            endpoint += f"&searchKeys[]=description&searchVals[]={description}"
        resp = self.session.get_json(endpoint)
        if df:
            return pd.DataFrame([m for m in resp["modelRepos"]])
        from predibase.resource_util import build_model_repo

        return [build_model_repo(m, self.session) for m in resp["modelRepos"]]

    def get_model_repo(self, name: str) -> ModelRepo:
        return get_model_repo(self.session, name)

    def create_model_repo(
        self,
        name: str,
        description: Optional[str] = None,
        engine: Optional[Engine] = None,
        exists_ok: bool = False,
    ) -> ModelRepo:
        return create_model_repo(
            self.session,
            name,
            description=description,
            engine=engine,
            exists_ok=exists_ok,
        )

    def delete_model_repo(self, name: Optional[str] = None, repo_id: Optional[int] = None):
        if name and repo_id:
            raise ValueError("Cannot provide both repo name and repo id")
        if name:
            repo = self.get_model_repo(name)
            repo_id = repo.id
        endpoint = f"/models/repo/{repo_id}"
        return self.session.delete_json(endpoint)

    def get_model(
        self,
        name: Optional[str] = None,
        version: Optional[Union[int, str]] = None,
        model_id: Optional[int] = None,
    ) -> Model:
        """
        :param name: name of model repo
        :param version: version in <repo_name>.<version>. If no version is specified, default to most recent
        ready/deployed/deploying/undeploying model
        :param model_id: Overrides repo name and version to get the exact model corresponding to the id
        :return:
        """
        return _get_model(self.session, name, version, model_id)

    def get_suggested_configs(
        self,
        base_config: Dict[Any, str],
    ) -> List[ConfigSuggestion]:
        return _get_suggested_configs(self.session, base_config)

    def get_most_recent_model_with_description(self, repo_name: str, model_description: str) -> Union[Model, None]:
        """Get the most recent model in the repo with a specific description.

        Returns None if a model with that description doesn't exist or if the repo doesn't exist.

        Args:
            repo_name: model repository name.
            model_description: model description.
        """
        try:
            repo = self.get_model_repo(repo_name)
        except ServerResponseError as e:
            if e.code == 400:
                # Model repo doesn't exist.
                return None

        models = repo.list_models()
        models.sort(key=lambda m: m.version, reverse=True)

        for model in models:
            if model.description == model_description:
                return model
        return None

    def get_default_config(
        self,
        dataset: Dataset,
        targets: Optional[Union[str, List[str]]] = None,
        automl: Optional[bool] = False,
        hyperopt: Optional[bool] = False,
    ) -> dict:
        """
        :param dataset
        :param targets: single target or list of targets. If targets are not provided, the user will
        have to manually edit config["output_features"] for training.
        :param hyperopt: whether hyperopt is enabled
        :return: Ludwig config
        """
        return _get_default_config(self.session, dataset, targets, automl, hyperopt)

    def create_model(
        self,
        repository_name: str,
        dataset: Dataset,
        config: dict,
        engine: Optional[Engine] = None,
        repo_description: Optional[str] = None,
        model_description: Optional[str] = None,
    ) -> Model:
        if not self.session.is_plan_expired():
            if isinstance(config, str):  # assume path
                config_dict = load_yaml(config)
                self.config_fp = config
            elif config:
                config_dict = copy.deepcopy(config)
                self.config_fp = None
            repo_id = self._create_or_retrieve_repository(repository_name, repo_description, engine)
            return self._create_and_train_model(
                repo_id=repo_id,
                config=config_dict,
                dataset=dataset,
                repository_name=repository_name,
                model_description=model_description,
                engine=engine,
            )
        else:
            raise PermissionError(
                "Training models is locked for expired plans. Contact us to upgrade.",
            )

    def train_suggested_models(
        self,
        repository_name: str,
        dataset: Dataset,
        config_suggestions: List[ConfigSuggestion],
        config_suggestions_mask: Optional[List[bool]] = None,
        engine: Optional[Engine] = None,
        repository_description: Optional[str] = "",
    ) -> List[Model]:
        """Train a number of models on list of config suggestions.

        Args:
            repository_name: name of the model repository.
            dataset: dataset object.
            config_suggestions: list of ConfigSuggestion objects.
            config_suggestions_mask: list of booleans referencing which
                config_suggestions to train on. Length must be equal to that
                of config_suggestions.
            engine: Engine object to use for training.
            repository_description: description of the repository.

        Returns:
            List of Models.
        """
        if not self.session.is_plan_expired():
            if config_suggestions_mask and len(config_suggestions_mask) != len(config_suggestions):
                raise ValueError("len(config_suggestions_mask) must be equal to len(config_suggestions).")
            if config_suggestions_mask and any([not isinstance(elem, bool) for elem in config_suggestions_mask]):
                raise ValueError("Elements of config_suggestions_mask must be all boolean.")
            if engine is not None and engine.template_id != 8:
                raise ValueError(
                    "Cannot train suggested models with this engine. Please specify a training (adaptive) engine.",
                )
            if engine is None:
                all_adaptive_engines = [eng for eng in self.list_engines() if eng.template_id == 8]
                if all_adaptive_engines:
                    engine = all_adaptive_engines[-1]
                    log_info(f"Using engine `{engine.name}` to train the suggested configs.")
                else:
                    raise ValueError(
                        "Tried to assign a training (adaptive) engine to train suggested models,"
                        "but found none. Please create a training engine.",
                    )
            if config_suggestions_mask:
                config_suggestions = [config_suggestions[i] for i, elem in enumerate(config_suggestions_mask) if elem]

            repo_id = self._create_or_retrieve_repository(repository_name, repository_description, engine)
            models = []
            with ThreadPoolExecutor() as executor:
                futures = [
                    executor.submit(
                        self._create_and_train_model,
                        repo_id,
                        suggestion.config,
                        dataset,
                        repository_name,
                        engine,
                        suggestion.description,
                    )
                    for suggestion in config_suggestions
                ]
                for future in as_completed(futures):
                    models.append(future.result())
            return models
        else:
            raise PermissionError(
                "Training models is locked for expired plans. Contact us to upgrade.",
            )

    def upload_model(self, zip_fp: str, repo: ModelRepo, dataset: Optional[Dataset] = None) -> Model:
        return repo.upload_model(zip_fp, dataset=dataset)

    def add_feature(self, config: dict, name: str, type: str, preprocessing: Optional[dict] = None):
        """This is a helper method for adding a feature to a model config.

        :param config: The config to add a feature to
        :param name: The name of the feature to add to the config
        :param type: The type of the feature to add to the config
        :param preprocessing: An optional preprocessing dictionary to specify preprocessing instructions for this
            particular feature.
        """
        if name in [x["name"] for x in config["input_features"]]:
            raise ValueError(f"Input Feature: '{name}' already exists")

        feature = {"name": name, "type": type}

        if preprocessing:
            feature["preprocessing"] = preprocessing

        config["input_features"].append(feature)

    def remove_feature(
        self,
        config: dict,
        name: str,
    ):
        """This is a helper method for removing a feature from a model config.

        :param config: The config to remove a feature from
        :param name: The name of the feature to remove from the config
        """
        if name not in [x["name"] for x in config["input_features"]]:
            raise ValueError(f"Input Feature: '{name}' does not exist")

        updated_features = [x for x in config["input_features"] if x["name"] != name]

        config["input_features"] = updated_features

    def _create_or_retrieve_repository(
        self,
        repo_name: str,
        repo_description: Optional[str] = None,
        engine: Optional[Engine] = None,
        exists_ok: bool = True,
    ) -> int:
        """Creates a new model repository or retrieves an existing one."""
        try:
            resp = self.session.post_json(
                "/models/repo",
                {
                    "modelType": "ludwig",
                    "description": repo_description,
                    "retrainCadence": "",
                    "retrainConfig": {},
                    "modelName": repo_name,
                    "selectedEngineID": engine.id if engine else None,
                    "parentID": None,
                },
            )
            log_info(f"Created model repository: <{repo_name}>")
            repo_id = resp["id"]
            return repo_id
        except ServerResponseError as e:
            if exists_ok and e.code == 400:
                log_info(f"Model repository {repo_name} already exists and new models will be added to it.")
                resp = self.session.get_json(f"/models/repo/name/{encode_url_param(repo_name)}")
                repo_id = resp["id"]
                return repo_id
            else:
                raise e

    def _create_and_train_model(
        self,
        repo_id: int,
        config: dict,
        dataset: Dataset,
        repository_name: str,
        engine: Optional[Engine] = None,
        model_description: Optional[str] = None,
    ):
        resp = self.session.post_json(
            "/models/train",
            {
                "datasetID": dataset.id,
                "config": ConfigEncoder().decimalize(config),
                "repoID": repo_id,
                "engineID": engine.id if engine else None,
                "description": model_description,
            },
        )
        model_id, model_version = resp["model"]["id"], resp["model"]["repoVersion"]
        log_info(f"Training model version {model_version} for model repository <{repository_name}>...")

        endpoint = f"models/version/{model_id}"
        url = get_url(self.session, endpoint)
        log_info(f"Check Status of Model Training Here: [link={url}]{url}")

        resp = self.session.get_json_until(
            f"/models/version/{model_id}",
            lambda resp: resp["modelVersion"]["completed"] is not None
            and resp["modelVersion"]["status"]
            not in (
                "failed",
                "canceled",
            ),
            lambda resp: (
                f"Failed to train model with status {resp['modelVersion']['status'].upper()} and error "
                f"{resp['modelVersion']['errorText']}"
            )
            if resp["modelVersion"]["status"] in {"failed", "canceled"}
            else None,
        )

        from predibase.resource_util import build_model

        return build_model(resp["modelVersion"], self.session)
